from ir_measures import measures
from .base import Measure, ParamInfo


class _nDCG(measures.Measure):
    """
    The normalized Discounted Cumulative Gain (nDCG).
    Uses graded labels - systems that put the highest graded documents at the top of the ranking.
    It is normalized wrt. the Ideal NDCG, i.e. documents ranked in descending order of graded label.

<cite>
@article{Jarvelin:2002:CGE:582415.582418,
  author = {J\"{a}rvelin, Kalervo and Kek\"{a}l\"{a}inen, Jaana},
  title = {Cumulated Gain-based Evaluation of IR Techniques},
  journal = {ACM Trans. Inf. Syst.},
  volume = {20},
  number = {4},
  year = {2002},
  pages = {422--446},
  numpages = {25},
  url = {http://doi.acm.org/10.1145/582415.582418},
}
</cite>
    """
    __name__ = 'nDCG'
    NAME = __name__
    PRETTY_NAME = 'Normalised Discounted Cumulative Gain'
    SHORT_DESC = 'A measure of the total gain a user encounters in a result list, discounted by rank and normalised against an ideal ranking.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'dcg': measures.ParamInfo(dtype=str, choices=['log2', 'exp-log2'], default='log2', desc='DCG formulation'),
        'gains': measures.ParamInfo(dtype=dict, desc='custom gain mapping (int-to-int)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='ignore returned documents that do not have relevance judgments'),
    }


nDCG = _nDCG()
NDCG = nDCG
measures.register(nDCG, ['NDCG'])



from ir_measures import measures
from .base import Measure, ParamInfo, SumAgg


class _INST(measures.Measure):
    """
    INST, a variant of INSQ

<cite>
@inproceedings{10.1145/2766462.2767728,
  author = {Bailey, Peter and Moffat, Alistair and Scholer, Falk and Thomas, Paul},
  title = {User Variability and IR System Evaluation},
  year = {2015},
  booktitle = {Proceedings of the 38th International ACM SIGIR Conference on Research and Development in Information Retrieval},
  pages = {625–634},
  series = {SIGIR '15},
  url = {https://doi.org/10.1145/2766462.2767728}
}
</cite>
    """
    __name__ = 'INST'
    NAME = __name__
    PRETTY_NAME = 'INST'
    SHORT_DESC = 'An improved version of INSQ that better handles when either no documents or all retrieved documents are relevant.'
    SUPPORTED_PARAMS = {
        'T': measures.ParamInfo(dtype=float, default=1.0, desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


class _INSQ(measures.Measure):
    """
    INSQ

<cite>
@inproceedings{Moffat:2012:MMI:2407085.2407092,
  author = {Moffat, Alistair and Scholer, Falk and Thomas, Paul},
  title = {Models and Metrics: IR Evaluation As a User Process},
  booktitle = {Proceedings of the Seventeenth Australasian Document Computing Symposium},
  year = {2012},
  url = {http://doi.acm.org/10.1145/2407085.2407092}
}
</cite>
    """
    __name__ = 'INSQ'
    NAME = __name__
    PRETTY_NAME = 'INSQ'
    SHORT_DESC = 'A weighted precision measure based on the conditional probability of the user continuing to the next item.'
    SUPPORTED_PARAMS = {
        'T': measures.ParamInfo(dtype=float, default=1.0, desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


INST = _INST()
measures.register(INST)

INSQ = _INSQ()
measures.register(INSQ)



import ir_measures
from ir_measures import providers, measures, Metric


class RuntimeProvider(providers.Provider):
    """
    Supports measures that are defined at runtime via `ir_measures.define()` and
    `ir_measures.define_byquery()`.
    """
    NAME = 'runtime'

    def supports(self, measure):
        measure.validate_params()
        if hasattr(measure, 'runtime_impl'):
            return True
        return False

    def _evaluator(self, measures, qrels):
        measures = ir_measures.util.flatten_measures(measures)
        # Convert qrels to dict_of_dict (input format used by pytrec_eval)
        qrels = ir_measures.util.QrelsConverter(qrels).as_pd_dataframe()
        qrels.sort_values(by=['query_id', 'doc_id'], inplace=True)
        return RuntimeEvaluator(measures, qrels)


class RuntimeEvaluator(providers.Evaluator):
    def __init__(self, measures, qrels):
        super().__init__(measures, set(qrels['query_id'].unique()))
        self.qrels = qrels

    def _iter_calc(self, run):
        run = ir_measures.util.RunConverter(run).as_pd_dataframe()
        run.sort_values(by=['query_id', 'score'], ascending=[True, False], inplace=True)
        for measure in self.measures:
            yield from measure.runtime_impl(self.qrels, run)


def define(impl, name=None, support_cutoff=True):
    _SUPPORTED_PARAMS = {}
    if support_cutoff:
        _SUPPORTED_PARAMS['cutoff'] = measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold')
    class _RuntimeMeasure(measures.Measure):
        nonlocal _SUPPORTED_PARAMS
        SUPPORTED_PARAMS = _SUPPORTED_PARAMS
        NAME = name
        __name__ = name

        def runtime_impl(self, qrels, run):
            if 'cutoff' in self.params and self.params['cutoff'] is not None:
                cutoff = self.params['cutoff']
                # assumes results already sorted (as is done in RuntimeEvaluator)
                run = run.groupby('query_id').head(cutoff).reset_index(drop=True)
            for qid, score in impl(qrels, run):
                yield Metric(qid, self, score)
    return _RuntimeMeasure()


def _byquery_impl(impl):
    def _wrapped(qrels, run):
        for qid, run_subdf in run.groupby("query_id"):
            qrels_subdf = qrels[qrels['query_id'] == qid]
            res = impl(qrels_subdf, run_subdf)
            yield qid, res
    return _wrapped


def define_byquery(impl, name=None, support_cutoff=True):
    return define(_byquery_impl(impl), name or repr(impl), support_cutoff)


providers.register(RuntimeProvider())



import unittest
import itertools
import ir_measures


class TestRuntime(unittest.TestCase):

    def test_define_byquery(self):
        def my_p(qrels, run):
            run = run.merge(qrels, 'left', on=['doc_id'])
            return (run['relevance'] > 0).sum() / len(run)
        def my_s(qrels, run):
            run = run.merge(qrels, 'left', on=['doc_id'])
            return 1. if (run['relevance'] > 0).sum() else 0.
        MyP = ir_measures.define_byquery(my_p)
        MyS = ir_measures.define_byquery(my_s)
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 0
1 0 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))

        result = list((MyP@1).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@1).calc_aggregate(qrels, run), 0.0)

        result = list((MyP@2).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.5)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@2).calc_aggregate(qrels, run), 0.25)

        result = list((MyP@3).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.6666666666666666)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@3).calc_aggregate(qrels, run), 0.3333333333333333)

        result = list((MyS@2).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyS@2).calc_aggregate(qrels, run), 0.5)

    def test_define(self):
        def my_p(qrels, run):
            run = run.merge(qrels, 'left', on=['query_id', 'doc_id'])
            for qid, df in run.groupby('query_id'):
                yield qid, (df['relevance'] > 0).sum() / len(df)
        def my_s(qrels, run):
            run = run.merge(qrels, 'left', on=['query_id', 'doc_id'])
            for qid, df in run.groupby('query_id'):
                yield qid, 1. if (df['relevance'] > 0).sum() else 0.
        MyP = ir_measures.define(my_p)
        MyS = ir_measures.define(my_s)
        qrels = list(ir_measures.read_trec_qrels('''
0 0 D0 0
0 0 D1 1
0 0 D2 1
0 0 D3 2
0 0 D4 0
1 0 D0 1
1 0 D3 0
1 0 D5 2
'''))
        run = list(ir_measures.read_trec_run('''
0 0 D0 1 0.8 run
0 0 D2 2 0.7 run
0 0 D1 3 0.3 run
0 0 D3 4 0.4 run
0 0 D4 5 0.1 run
1 0 D1 1 0.8 run
1 0 D3 2 0.7 run
1 0 D4 3 0.3 run
1 0 D2 4 0.4 run
'''))
        result = list((MyP@1).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@1).calc_aggregate(qrels, run), 0.0)

        result = list((MyP@2).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.5)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@2).calc_aggregate(qrels, run), 0.25)

        result = list((MyP@3).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 0.6666666666666666)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyP@3).calc_aggregate(qrels, run), 0.3333333333333333)

        result = list((MyS@2).iter_calc(qrels, run))
        self.assertEqual(result[0].query_id, "0")
        self.assertEqual(result[0].value, 1.)
        self.assertEqual(result[1].query_id, "1")
        self.assertEqual(result[1].value, 0.)
        self.assertEqual((MyS@2).calc_aggregate(qrels, run), 0.5)


if __name__ == '__main__':
    unittest.main()



from ir_measures import measures
from .base import Measure, ParamInfo


class _Compat(measures.Measure):
    """
    Compatibility measure desribed in:

<cite>
@article{10.1145/3451161,
  author = {Clarke, Charles L. A. and Vtyurina, Alexandra and Smucker, Mark D.},
  title = {Assessing Top-k Preferences},
  journal = {ACM Transactions on Information Systems},
  volume = {39},
  number = {3},
  articleno = {33},
  numpages = {21},
  year = {2021},
  url = {https://doi.org/10.1145/3451161},
}
</cite>
    """
    __name__ = 'Compat'
    NAME = __name__
    PRETTY_NAME = 'Compatibility'
    SHORT_DESC = 'The Rank Biased Overlap between the results and an ideal ranking.'
    SUPPORTED_PARAMS = {
        'p': measures.ParamInfo(dtype=float, default=0.95, desc='persistence'),
        'normalize': measures.ParamInfo(dtype=bool, default=True, desc='apply normalization for finite ideal rankings'),
    }


Compat = _Compat()
measures.register(Compat)



from ir_measures import measures
from .base import Measure, ParamInfo


class _BPM(measures.Measure):
    """
    The Bejeweled Player Model (BPM).

<cite>
@inproceedings{Zhang:2017:EWS:3077136.3080841,
  author = {Zhang, Fan and Liu, Yiqun and Li, Xin and Zhang, Min and Xu, Yinghui and Ma, Shaoping},
  title = {Evaluating Web Search with a Bejeweled Player Model},
  booktitle = {SIGIR},
  year = {2017},
  url = {http://doi.acm.org/10.1145/3077136.3080841}
}
</cite>
    """
    __name__ = 'BPM'
    NAME = __name__
    PRETTY_NAME = 'Bejeweled Player Model'
    SHORT_DESC = 'A measure that balances both gain and user patience to determine when they stop traversing search results.'
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'T': measures.ParamInfo(dtype=float, default=1., desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


BPM = _BPM()
measures.register(BPM)
